import os
import subprocess
from pathlib import Path

import pytest
import yaml
from click.testing import CliRunner
from kedro.context import KedroContext, load_context


@pytest.fixture
def cli_runner():
    runner = CliRunner()
    with runner.isolated_filesystem():
        yield runner


@pytest.fixture
def dummy_kedro_config(cli_runner):
    config = {
        "project_name": "Dummy project",
        "repo_name": "dummy-project",
        "python_package": "dummy_project",
        "include_example": True,
        "output_dir": os.getcwd(),
    }

    config_path = Path("config.yml")

    with config_path.open("w", encoding="utf-8") as file_:
        yaml.dump(config, file_)

    return config_path.resolve()


@pytest.fixture
def dummy_kedro_project(dummy_kedro_config):
    cwd_ = os.getcwd()
    subprocess.check_call(["kedro", "new", "-c", str(dummy_kedro_config)])

    project_dir = Path("dummy-project").resolve()
    os.chdir(str(project_dir))
    yield project_dir
    os.chdir(cwd_)


@pytest.fixture
def project_context(dummy_kedro_project) -> KedroContext:
    return load_context(".")



"""Kedro run manager implementation: ProcessManager"""
import abc
import sys
import uuid
from functools import wraps
from multiprocessing import Process, Queue
from queue import Empty
from typing import Any, AnyStr, Callable, Dict, List, Union


def _get_new_events(events_queue: Queue):
    while True:
        try:
            yield events_queue.get_nowait()
        except Empty:
            break


def _wrapped_write(
    queue: Queue, real_write: Callable[[AnyStr], int]
) -> Callable[[AnyStr], int]:
    """Forward messages sent to `real_write` to a multiprocessing queue.

    Args:
        queue: Multiprocessing queue to forward the messages to.
        real_write: Write callable the message was addressed to. It's signature
            must be identical to `sys.stdout.write`.

    Returns:
        Wrapped write function.
    """

    # the signature of `_wrapped` is deliberately kept as close
    # to sys.stdout.write as possible
    @wraps(real_write)
    def _wrapped(s: AnyStr) -> int:  # pylint: disable=invalid-name
        s_to_queue = s.decode("utf-8") if isinstance(s, bytes) else s  # type: ignore
        if s_to_queue.strip():
            queue.put(s_to_queue)
        real_write(s)
        return len(s)

    return _wrapped


class AbstractManager(abc.ABC):
    """`AbstractManager` is the base class for all pipeline run managers
    """

    def __init__(
        self,
        context: Any,
        run_id: str = None,
        run_args: dict = None,
        extra_params: dict = None,
    ):
        """
        Instantiates the run manager class
        :param context: Project context
        :param run_id: Specific Run ID
        :param run_args: Run args
        :param extra_params: Extra params
        """
        self._context = context
        self._run_id = run_id or str(uuid.uuid4())
        self._run_args = run_args
        self._extra_params = extra_params or {}
        self._events = []  # type: List[str]
        self._run_finished = False

    @property
    def run_id(self):
        """Run_ID getter"""
        return self._run_id

    @property
    def events(self):
        """Events getter"""
        return self._events

    @abc.abstractmethod
    def start(self):
        """The abstract interface for starting run managers"""
        raise NotImplementedError(
            "`{}` is a subclass of AbstractManager and"
            "it must implement the `start` method".format(self.__class__.__name__)
        )

    @abc.abstractmethod
    def stop(self):
        """The abstract interface for stopping run managers"""
        raise NotImplementedError(
            "`{}` is a subclass of AbstractManager and"
            "it must implement the `stop` method".format(self.__class__.__name__)
        )

    @abc.abstractmethod
    def status(self):
        """The abstract interface for getting status of runs"""
        raise NotImplementedError(
            "`{}` is a subclass of AbstractManager and"
            "it must implement the `status` method".format(self.__class__.__name__)
        )


class ProcessManager(AbstractManager):
    """ProcessManager is an AbstractManager implementation.
    Uses multiprocessing Process to manage kedro runs"""

    def __init__(
        self,
        context,
        run_id=None,
        proc=None,
        queue=None,
        run_args=None,
        extra_params=None,
    ):
        """
        Instantiates the run manager class
        """
        super().__init__(
            context=context,
            run_id=run_id,
            run_args=run_args,
            extra_params=extra_params,
        )
        self._proc = proc or None
        self._proc_queue = queue or Queue()  # type: Queue

    @property
    def proc(self) -> Process:
        """Process getter"""
        return self._proc

    @property
    def proc_queue(self):
        """Process queue getter"""
        return self._proc_queue

    def start(self):
        """
        Start run process with queue
        """
        self._proc = Process(target=self._wrapped_run, daemon=True)
        self._proc.start()

    def stop(self) -> str:
        """
        Stop current process
        :return: Stop message
        """
        raise NotImplementedError("Run stop is not supported yet.")

    def status(self) -> Dict[Any, Union[str, list]]:
        """
        Return status of the current process
        :return:
        """

        new_events = _get_new_events(self._proc_queue)
        self.events.extend(new_events)
        return dict(
            run_status="Pending" if self.proc.is_alive() else "Completed",
            events=self._events,
        )

    def _wrapped_run(self):
        """Enhanced pipeline run to collect events"""
        sys.stdout.write = _wrapped_write(self._proc_queue, sys.stdout.write)
        sys.stderr.write = _wrapped_write(self._proc_queue, sys.stderr.write)

        self._proc_queue.put("Starting run")
        self._context.run(**self._run_args)
        self._proc_queue.put("Completed run")



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: kedro_grpc_server/kedro.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='kedro_grpc_server/kedro.proto',
  package='kedro',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1dkedro_grpc_server/kedro.proto\x12\x05kedro\"-\n\nRunSummary\x12\x0e\n\x06run_id\x18\x01 \x01(\t\x12\x0f\n\x07success\x18\x02 \x01(\t\"0\n\tRunParams\x12\x15\n\rpipeline_name\x18\x01 \x01(\t\x12\x0c\n\x04tags\x18\x02 \x01(\t\"#\n\x0fPipelineSummary\x12\x10\n\x08pipeline\x18\x01 \x03(\t\"\x10\n\x0ePipelineParams\"\x17\n\x05RunId\x12\x0e\n\x06run_id\x18\x01 \x01(\t\"c\n\tRunStatus\x12\x0e\n\x06\x65vents\x18\x01 \x03(\t\x12\x11\n\texit_code\x18\x02 \x01(\t\x12\x0e\n\x06run_id\x18\x03 \x01(\t\x12\x0f\n\x07success\x18\x04 \x01(\t\x12\x12\n\nrun_status\x18\x05 \x01(\t2\xa1\x01\n\x05Kedro\x12>\n\rListPipelines\x12\x15.kedro.PipelineParams\x1a\x16.kedro.PipelineSummary\x12*\n\x03Run\x12\x10.kedro.RunParams\x1a\x11.kedro.RunSummary\x12,\n\x06Status\x12\x0c.kedro.RunId\x1a\x10.kedro.RunStatus\"\x00\x30\x01\x62\x06proto3'
)




_RUNSUMMARY = _descriptor.Descriptor(
  name='RunSummary',
  full_name='kedro.RunSummary',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='run_id', full_name='kedro.RunSummary.run_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='success', full_name='kedro.RunSummary.success', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=40,
  serialized_end=85,
)


_RUNPARAMS = _descriptor.Descriptor(
  name='RunParams',
  full_name='kedro.RunParams',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='pipeline_name', full_name='kedro.RunParams.pipeline_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='tags', full_name='kedro.RunParams.tags', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=87,
  serialized_end=135,
)


_PIPELINESUMMARY = _descriptor.Descriptor(
  name='PipelineSummary',
  full_name='kedro.PipelineSummary',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='pipeline', full_name='kedro.PipelineSummary.pipeline', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=137,
  serialized_end=172,
)


_PIPELINEPARAMS = _descriptor.Descriptor(
  name='PipelineParams',
  full_name='kedro.PipelineParams',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=174,
  serialized_end=190,
)


_RUNID = _descriptor.Descriptor(
  name='RunId',
  full_name='kedro.RunId',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='run_id', full_name='kedro.RunId.run_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=192,
  serialized_end=215,
)


_RUNSTATUS = _descriptor.Descriptor(
  name='RunStatus',
  full_name='kedro.RunStatus',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='events', full_name='kedro.RunStatus.events', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='exit_code', full_name='kedro.RunStatus.exit_code', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='run_id', full_name='kedro.RunStatus.run_id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='success', full_name='kedro.RunStatus.success', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='run_status', full_name='kedro.RunStatus.run_status', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=217,
  serialized_end=316,
)

DESCRIPTOR.message_types_by_name['RunSummary'] = _RUNSUMMARY
DESCRIPTOR.message_types_by_name['RunParams'] = _RUNPARAMS
DESCRIPTOR.message_types_by_name['PipelineSummary'] = _PIPELINESUMMARY
DESCRIPTOR.message_types_by_name['PipelineParams'] = _PIPELINEPARAMS
DESCRIPTOR.message_types_by_name['RunId'] = _RUNID
DESCRIPTOR.message_types_by_name['RunStatus'] = _RUNSTATUS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

RunSummary = _reflection.GeneratedProtocolMessageType('RunSummary', (_message.Message,), {
  'DESCRIPTOR' : _RUNSUMMARY,
  '__module__' : 'kedro_grpc_server.kedro_pb2'
  # @@protoc_insertion_point(class_scope:kedro.RunSummary)
  })
_sym_db.RegisterMessage(RunSummary)

RunParams = _reflection.GeneratedProtocolMessageType('RunParams', (_message.Message,), {
  'DESCRIPTOR' : _RUNPARAMS,
  '__module__' : 'kedro_grpc_server.kedro_pb2'
  # @@protoc_insertion_point(class_scope:kedro.RunParams)
  })
_sym_db.RegisterMessage(RunParams)

PipelineSummary = _reflection.GeneratedProtocolMessageType('PipelineSummary', (_message.Message,), {
  'DESCRIPTOR' : _PIPELINESUMMARY,
  '__module__' : 'kedro_grpc_server.kedro_pb2'
  # @@protoc_insertion_point(class_scope:kedro.PipelineSummary)
  })
_sym_db.RegisterMessage(PipelineSummary)

PipelineParams = _reflection.GeneratedProtocolMessageType('PipelineParams', (_message.Message,), {
  'DESCRIPTOR' : _PIPELINEPARAMS,
  '__module__' : 'kedro_grpc_server.kedro_pb2'
  # @@protoc_insertion_point(class_scope:kedro.PipelineParams)
  })
_sym_db.RegisterMessage(PipelineParams)

RunId = _reflection.GeneratedProtocolMessageType('RunId', (_message.Message,), {
  'DESCRIPTOR' : _RUNID,
  '__module__' : 'kedro_grpc_server.kedro_pb2'
  # @@protoc_insertion_point(class_scope:kedro.RunId)
  })
_sym_db.RegisterMessage(RunId)

RunStatus = _reflection.GeneratedProtocolMessageType('RunStatus', (_message.Message,), {
  'DESCRIPTOR' : _RUNSTATUS,
  '__module__' : 'kedro_grpc_server.kedro_pb2'
  # @@protoc_insertion_point(class_scope:kedro.RunStatus)
  })
_sym_db.RegisterMessage(RunStatus)



_KEDRO = _descriptor.ServiceDescriptor(
  name='Kedro',
  full_name='kedro.Kedro',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=319,
  serialized_end=480,
  methods=[
  _descriptor.MethodDescriptor(
    name='ListPipelines',
    full_name='kedro.Kedro.ListPipelines',
    index=0,
    containing_service=None,
    input_type=_PIPELINEPARAMS,
    output_type=_PIPELINESUMMARY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Run',
    full_name='kedro.Kedro.Run',
    index=1,
    containing_service=None,
    input_type=_RUNPARAMS,
    output_type=_RUNSUMMARY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Status',
    full_name='kedro.Kedro.Status',
    index=2,
    containing_service=None,
    input_type=_RUNID,
    output_type=_RUNSTATUS,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_KEDRO)

DESCRIPTOR.services_by_name['Kedro'] = _KEDRO

# @@protoc_insertion_point(module_scope)



"""Kedro gRPC Server"""
import logging
import time
from concurrent import futures
from copy import deepcopy
from typing import Any

import grpc
from kedro.framework.cli import get_project_context

from kedro_grpc_server.kedro_pb2 import (  # type: ignore
    PipelineSummary,
    RunStatus,
    RunSummary,
)
from kedro_grpc_server.kedro_pb2_grpc import (  # type: ignore
    KedroServicer,
    add_KedroServicer_to_server,
)
from kedro_grpc_server.process_manager import ProcessManager


RUN_STATES = {}


class KedroServer(KedroServicer):
    """
    KedroServer is an implementation of KedroServicer
    """

    def __init__(self, context):
        self.app_context = context

    def ListPipelines(self, request, context):
        response = PipelineSummary()
        pipeline_names = self.app_context.pipelines.keys()
        response.pipeline.extend(pipeline_names)  # pylint: disable=no-member
        return response

    def Run(self, request, context):
        run_args = dict(pipeline_name=request.pipeline_name, tags=request.tags,)

        context = deepcopy(self.app_context)
        # run_id, run_states = start_run_process(
        #     context=context, run_args=run_args, extra_params={}
        # )
        #
        # RUN_STATES[run_id] = run_states
        proc_manager = ProcessManager(
            context=context, run_args=run_args, extra_params={}
        )
        run_id = proc_manager.run_id
        proc_manager.start()

        RUN_STATES[run_id] = proc_manager

        response = RunSummary()
        response.run_id = run_id
        response.success = f"Run {run_id} dispatched"
        return response

    def Status(self, request, context):
        """Get run status and logged events"""
        run_id = request.run_id
        response = RunStatus()

        if run_id not in RUN_STATES:
            response.events.extend([])  # pylint: disable=no-member
            response.run_status = "Error"
            response.exit_code = ""
            response.success = "Run ID doesn't exist"
            response.run_id = run_id
            yield response
        else:
            has_events = True

            process_info = RUN_STATES[run_id]  # type: ProcessManager

            while has_events:
                proc_status = process_info.status()
                run_status = proc_status.get("run_status")
                response.run_id = run_id
                response.events.extend(  # pylint: disable=no-member
                    proc_status.get("events")
                )
                response.success = "Status check was performed successfully"
                response.run_status = run_status
                response.exit_code = str(process_info.proc.exitcode)

                yield response
                has_events = bool(process_info.proc.is_alive())
                time.sleep(1)


class KedroGrpcServerException(Exception):
    """
    Raise an Kedro gRPC Server exception
    :raises Exception
    """

    pass


def grpc_serve(
    context: Any = None,
    host: str = "[::]",
    port: int = 50051,
    max_workers: int = 10,
    wait_term: bool = True,
):
    """
    Start the Kedro gRPC server

    :param context: Kedro Project Context
    :param host: host for running the grpc server
    :param port: Port to run the gRPC server on
    :param max_workers: Max number of workers
    :param wait_term: Wait for termination

    :raises KedroGrpcServerException: Failing to start gRPC Server
    """
    try:
        if not context:
            context = get_project_context()
        server = grpc.server(futures.ThreadPoolExecutor(max_workers=max_workers))
        add_KedroServicer_to_server(KedroServer(context), server)
        server.add_insecure_port(f"{host}:{port}")
        server.start()
        logging.info("Kedro gRPC Server started on %s", port)
        if wait_term:  # pragma: no cover
            server.wait_for_termination()
    except Exception as exc:
        logging.error(exc)
        raise KedroGrpcServerException("Failed to start Kedro gRPC Server")



import logging
import os
import signal
import sys
import time
from multiprocessing import Lock, Value
from typing import Dict

import grpc
import pytest
from kedro import __version__
from kedro.context import KedroContext
from kedro.io import DataCatalog
from kedro.pipeline import Pipeline, node
from kedro.versioning import Journal

from kedro_grpc_server import process_manager
from kedro_grpc_server.grpc_server import (  # type: ignore
    RUN_STATES,
    KedroGrpcServerException,
    KedroServer,
    grpc_serve,
)
from kedro_grpc_server.kedro_pb2 import RunId, RunParams  # type: ignore
from kedro_grpc_server.kedro_pb2_grpc import add_KedroServicer_to_server  # type: ignore
from kedro_grpc_server.process_manager import ProcessManager

_lock = Lock()  # pylint: disable=invalid-name
_num_active_runs = Value("i", 0)


def dummy_node():  # pragma: no cover
    return "X"


def bad_node():  # pragma: no cover
    raise ValueError("Oh no!!!")


def printing_node():  # pragma: no cover
    print("Printed from printing_node")
    with _lock:
        logging.info("Logged from printing_node")
        logging.error("Error from printing_node")
        return "X"


#
#
# @pytest.fixture
# def fake_run_id(mocker):
#     mocker.patch("kedro_server.utils._generate_run_id", return_value="abc123")


@pytest.fixture
def fake_run_id(mocker):
    def get_proc_manager():  # pylint: disable=unused-argument
        return ProcessManager(context=None, run_id="abc123")

    mocker.patch(
        "kedro_grpc_server.process_manager.ProcessManager.run_id",
        new=get_proc_manager().run_id,
    )


@pytest.fixture(autouse=True)
def clear_stdout_stderr_monkeypatch():
    stdout_write = sys.stdout.write
    stderr_write = sys.stderr.write
    yield
    # clear all monkeypatches after the test
    sys.stdout.write = stdout_write
    sys.stderr.write = stderr_write


class DummyContext(KedroContext):
    project_name = "test"
    project_version = __version__

    def _setup_logging(self) -> None:
        pass

    def _get_pipelines(self) -> Dict[str, Pipeline]:
        return {
            "__default__": Pipeline([node(dummy_node, None, "y")]),
            "my_pipeline": Pipeline([node(dummy_node, None, "y")]),
            "error_pipeline": Pipeline([node(bad_node, None, "empty")]),
            "printing_pipeline": Pipeline([node(printing_node, None, "y")]),
        }

    def _get_catalog(
        self,
        save_version: str = None,
        journal: Journal = None,
        load_versions: Dict[str, str] = None,
    ) -> DataCatalog:
        return DataCatalog()  # pragma: no cover

    def run(self, *args, **kwargs):  # pylint: disable=arguments-differ
        with _num_active_runs.get_lock():
            _num_active_runs.value += 1
        try:
            return super().run(*args, **kwargs)
        finally:
            with _num_active_runs.get_lock():
                _num_active_runs.value -= 1


@pytest.fixture(scope="module")
def grpc_add_to_server():
    return add_KedroServicer_to_server


@pytest.fixture(scope="module")
def grpc_servicer(tmpdir_factory):
    dummy_context = DummyContext(str(tmpdir_factory))

    return KedroServer(dummy_context)


@pytest.fixture(scope="module")
def grpc_stub(grpc_channel):
    from kedro_grpc_server.kedro_pb2_grpc import KedroStub

    return KedroStub(grpc_channel)


def grpc_server_on(channel, timeout=2) -> bool:
    try:
        grpc.channel_ready_future(channel).result(timeout=timeout)
        return True
    except grpc.FutureTimeoutError:
        return False


def test_grpc_server_on(grpc_channel):
    assert grpc_server_on(grpc_channel, 2)


def test_grpc_server_on_timeout():
    ch = grpc.insecure_channel("localhost:8081")

    assert not grpc_server_on(ch, 1)


def test_grpc_serve(tmpdir_factory):
    dummy_context = DummyContext(str(tmpdir_factory))
    grpc_serve(dummy_context, wait_term=False)

    ch = grpc.insecure_channel("localhost:50051")

    assert grpc_server_on(ch)


def test_grpc_serve_no_context(tmpdir_factory):
    with pytest.raises(KedroGrpcServerException) as exc:
        grpc_serve(wait_term=False)

    assert "Failed to start" in str(exc.value)


def test_get_pipelines(grpc_stub):
    from kedro_grpc_server.kedro_pb2 import PipelineParams

    request = PipelineParams()
    response = grpc_stub.ListPipelines(request)
    assert response.pipeline == [
        "__default__",
        "my_pipeline",
        "error_pipeline",
        "printing_pipeline",
    ]


@pytest.mark.usefixtures("fake_run_id")
def test_run_success(grpc_stub):
    expected = {"success": "Run abc123 dispatched", "run_id": "abc123"}
    request = RunParams()
    response = grpc_stub.Run(request)

    assert response.success == expected["success"]
    assert response.run_id == expected["run_id"]

    process = RUN_STATES[response.run_id].proc

    os.kill(process.pid, signal.SIGTERM)
    for _ in range(10):
        time.sleep(1)
        if not process.is_alive():
            print("alive")
            break
    else:
        process.terminate()  # pragma: no cover

    # clean-up
    process.join()


def test_get_status(grpc_stub):
    """Test to check that events logged or printed end up in
    the events queue and returned in the response, along with
    the correct run_status.
    """
    with _lock:
        request = RunParams()
        run_response = grpc_stub.Run(request)
        run_id = run_response.run_id

        status_request = RunId(run_id=run_id)
        status_response = grpc_stub.Status(status_request)

        for status in status_response:
            assert status.run_id == run_id
            assert status.run_status != ""

    RUN_STATES[run_id].proc.join(3)  # join the process to make sure it finished
    status_request = RunId(run_id=run_id)
    status_response = grpc_stub.Status(status_request)
    run_status = None
    events = []
    for status in status_response:
        run_status = status.run_status
        events = status.events
    assert run_status == "Completed"
    assert any("Completed run" in ev for ev in events)


def test_get_status_wrong_run_id(grpc_stub):
    status_request = RunId(run_id="invalid")
    status_response = grpc_stub.Status(status_request)
    for status in status_response:
        assert status.run_id == "invalid"
        assert status.success == "Run ID doesn't exist"
        assert status.run_status == "Error"


def test_wrapped_run(mocker, capsys):
    # need to assign both to variables as will be wrapped by '_wrapped_write' later
    stdout_write = sys.stdout.write
    stderr_write = sys.stderr.write
    # from kedro_server import process_manager

    def _fake_run(**kwargs):
        print("Fake stdout")
        print(f"Running: {kwargs}")
        print("Fake stderr", file=sys.stderr)

    mock_wrapped_write = mocker.spy(process_manager, "_wrapped_write")
    queue = mocker.Mock()
    context = mocker.Mock()
    context.run.side_effect = _fake_run

    fake_run_args = {"some_arg": "some_value"}
    proc_manager = ProcessManager(context=context, queue=queue, run_args=fake_run_args)
    proc_manager._wrapped_run()

    assert mock_wrapped_write.mock_calls == [
        mocker.call(proc_manager.proc_queue, stdout_write),  # patch of sys.stdout.write
        mocker.call(proc_manager.proc_queue, stderr_write),  # patch of sys.stderr.write
    ]

    assert queue.put.mock_calls == [
        mocker.call("Starting run"),
        mocker.call("Fake stdout"),
        mocker.call(f"Running: {fake_run_args}"),
        mocker.call("Fake stderr"),
        mocker.call("Completed run"),
    ]
    context.run.assert_called_once_with(**fake_run_args)

    captured = capsys.readouterr()  # capture what was sent to sys.stdout/stderr
    assert captured.out == f"Fake stdout\nRunning: {fake_run_args}\n"
    assert captured.err == "Fake stderr\n"



"""This is a Kedro plugin that creates a gRPC server for your kedro pipelines."""

from typing import Dict

import click

from kedro_grpc_server.grpc_server import grpc_serve
from kedro_grpc_server.process_manager import ProcessManager

HOST_HELP = """Host which the server will listen to. Defaults to 127.0.0.1."""
PORT_HELP = """TCP port which the server will listen to. Defaults to 4141."""
MAX_WORKERS_HELP = """Number of ThreadExecutors to handle RPCs."""

RUN_STATES = {}  # type: Dict[str, ProcessManager]


@click.group(name="Server")
def commands():
    """Kedro plugin for gRPC Server"""


@commands.group(name="server")
def server_commands():
    """Run project with Kedro-gRPC-Server"""


@server_commands.command()
@click.option("--host", default="127.0.0.1", help=HOST_HELP)
@click.option("--port", default=50051, type=int, help=PORT_HELP)
@click.option("--max_workers", default=10, type=int, help=MAX_WORKERS_HELP)
def grpc_start(host, port, max_workers, wait_term=True):
    """Start Kedro gRPC Server"""
    grpc_serve(
        host=host, port=port, max_workers=max_workers, wait_term=wait_term
    )  # pragma: no cover



import re
from os import path

from setuptools import find_packages, setup

name = "kedro_grpc_server"
here = path.abspath(path.dirname(__file__))

# get package version
with open(path.join(here, name, "__init__.py"), encoding="utf-8") as f:
    result = re.search(r'__version__ = ["\']([^"\']+)', f.read())
    if not result:
        raise ValueError("Can't find the version in kedro_grpc_server/__init__.py")
    version = result.group(1)

# get the dependencies and installs
with open("requirements.txt", "r", encoding="utf-8") as f:
    requires = [x.strip() for x in f if x.strip()]

# get test dependencies and installs
with open("test_requirements.txt", "r", encoding="utf-8") as f:
    test_requires = [x.strip() for x in f if x.strip() and not x.startswith("-r")]

# Get the long description from the README file
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    readme = f.read()

setup(
    name=name,
    version=version,
    description="Kedro gRPC Server, a Kedro plugin that creates a gRPC server for your kedro pipelines",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/mmchougule/kedro-grpc-server",
    author="Mayur Chougule",
    packages=find_packages(),
    python_requires=">=3.6, <3.8",
    include_package_data=True,
    tests_require=test_requires,
    install_requires=requires,
    zip_safe=False,
    entry_points={
        "kedro.project_commands": ["kedro_grpc_server = kedro_grpc_server.app:commands"]
    },
)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from kedro_grpc_server import kedro_pb2 as kedro__grpc__server_dot_kedro__pb2


class KedroStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.ListPipelines = channel.unary_unary(
                '/kedro.Kedro/ListPipelines',
                request_serializer=kedro__grpc__server_dot_kedro__pb2.PipelineParams.SerializeToString,
                response_deserializer=kedro__grpc__server_dot_kedro__pb2.PipelineSummary.FromString,
                )
        self.Run = channel.unary_unary(
                '/kedro.Kedro/Run',
                request_serializer=kedro__grpc__server_dot_kedro__pb2.RunParams.SerializeToString,
                response_deserializer=kedro__grpc__server_dot_kedro__pb2.RunSummary.FromString,
                )
        self.Status = channel.unary_stream(
                '/kedro.Kedro/Status',
                request_serializer=kedro__grpc__server_dot_kedro__pb2.RunId.SerializeToString,
                response_deserializer=kedro__grpc__server_dot_kedro__pb2.RunStatus.FromString,
                )


class KedroServicer(object):
    """Missing associated documentation comment in .proto file."""

    def ListPipelines(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Run(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Status(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_KedroServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'ListPipelines': grpc.unary_unary_rpc_method_handler(
                    servicer.ListPipelines,
                    request_deserializer=kedro__grpc__server_dot_kedro__pb2.PipelineParams.FromString,
                    response_serializer=kedro__grpc__server_dot_kedro__pb2.PipelineSummary.SerializeToString,
            ),
            'Run': grpc.unary_unary_rpc_method_handler(
                    servicer.Run,
                    request_deserializer=kedro__grpc__server_dot_kedro__pb2.RunParams.FromString,
                    response_serializer=kedro__grpc__server_dot_kedro__pb2.RunSummary.SerializeToString,
            ),
            'Status': grpc.unary_stream_rpc_method_handler(
                    servicer.Status,
                    request_deserializer=kedro__grpc__server_dot_kedro__pb2.RunId.FromString,
                    response_serializer=kedro__grpc__server_dot_kedro__pb2.RunStatus.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'kedro.Kedro', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class Kedro(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def ListPipelines(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/kedro.Kedro/ListPipelines',
            kedro__grpc__server_dot_kedro__pb2.PipelineParams.SerializeToString,
            kedro__grpc__server_dot_kedro__pb2.PipelineSummary.FromString,
            options, channel_credentials,
            call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Run(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/kedro.Kedro/Run',
            kedro__grpc__server_dot_kedro__pb2.RunParams.SerializeToString,
            kedro__grpc__server_dot_kedro__pb2.RunSummary.FromString,
            options, channel_credentials,
            call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Status(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/kedro.Kedro/Status',
            kedro__grpc__server_dot_kedro__pb2.RunId.SerializeToString,
            kedro__grpc__server_dot_kedro__pb2.RunStatus.FromString,
            options, channel_credentials,
            call_credentials, compression, wait_for_ready, timeout, metadata)
